/*=============================================================================
libxlg
base/filter.c - Control functions for picture filters
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: filter.c,v 1.8 2004/04/15 13:54:40 root Exp $
=============================================================================*/
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xlg.h"
#include "filters/strap.h"

static void strap_filters(void);

static pthread_mutex_t strap_mutex = PTHREAD_MUTEX_INITIALIZER;
static int strapped = 0;

static pthread_mutex_t filter_list_mutex = PTHREAD_MUTEX_INITIALIZER;
static XLG_filter *filter_list = NULL;
static size_t num_filters = 0;

//-----------------------------------------------------------------------------
// Not static since external SO/DLLs might need it
/* static */ int xlg_flt_register(XLG_filter *fsp) {
    /* Registers a filter within the xlg_flt database, which can later be
    retrieved using xlg_flt_list(). A filter does not need to be enregistered,
    it can be used independently. xlg_flt_register() returns -errno upon error,
    0 if the filter is already registered, and >0 for success. */
    XLG_filter *lp = NULL, *travp;

    pthread_mutex_lock(&filter_list_mutex);
    travp = (XLG_filter *)&filter_list;

    while(travp->Next != NULL) {
        // Alphabetic sorting
        int v = strcmp(fsp->name, travp->Next->name);
        if(v == 0) {
            pthread_mutex_unlock(&filter_list_mutex);
            return 0;
        }
        if(v < 0) {
            lp = travp->Next;
            break;
        }
        travp = travp->Next;
    }

    travp->Next = fsp;
    fsp->Next = lp;
    ++num_filters;
    pthread_mutex_unlock(&filter_list_mutex);
    return 1;
}

// Not static since external SO/DLLs might need it
/* static */ int xlg_flt_unregister(XLG_filter *fsp) {
    /* Unregisters a filter. Returns 0 if there was no such filter in the list,
    >0 if unregistering succeeded, and -errno for an error. */
    XLG_filter *travp;

    pthread_mutex_lock(&filter_list_mutex);
    travp = (XLG_filter *)&filter_list;
    while(travp->Next != NULL) {
        if(strcmp(travp->Next->name, fsp->name) == 0) {
            travp->Next = fsp->Next;
            --num_filters;
            pthread_mutex_unlock(&filter_list_mutex);
            fsp->Next = NULL;
            return 1;
        }
        travp = travp->Next;
    }
    return 0;
}

int xlg_flt_list(XLG_filter ***p) {
    XLG_filter *travp, **area, **ap;
    int n;

    if(!strapped) { strap_filters(); }
    pthread_mutex_lock(&filter_list_mutex);
    if(num_filters == 0) { return 0; }

    if((area = ap = malloc(sizeof(XLG_filter *) * num_filters)) == NULL) {
        return -errno;
    }

    travp = filter_list;
    n = num_filters;
    while(n--) {
        *ap++ = travp;
        travp = travp->Next;
    }

    pthread_mutex_unlock(&filter_list_mutex);
    *p = area;
    return num_filters;
}

XLG_filter *xlg_flt_get(const char *nm) {
    XLG_filter *travp;

    if(nm == NULL) { return NULL; }
    if(!strapped) { strap_filters(); }
    pthread_mutex_lock(&filter_list_mutex);
    travp = filter_list;

    while(travp != NULL) {
        if(strcmp(travp->name, nm) == 0) {
            pthread_mutex_unlock(&filter_list_mutex);
            return travp;
        }
        travp = travp->Next;
    }

    pthread_mutex_unlock(&filter_list_mutex);
    return NULL;
}

//-----------------------------------------------------------------------------
#define FLT(__name, __desc, __proc, __proc_gs, __show) \
 {NULL, (__name), (__desc), (__proc), (__proc_gs), (__show)}
static XLG_filter fltinfo[] = {
    FLT("copy",        "NOOP filter",                          xlg_flt_copy,        NULL,              0),
    FLT("cquant",      "Color quantization (bits, Grayscale)", xlg_flt_cquant,      NULL,              1),
    FLT("dft",         "Discrete Fourier Transform",           xlg_flt_dft,         xlg_flt_dft_GS,    1),
    FLT("flip",        "Flip: Vertical",                       xlg_flt_flip,        NULL,              1),
    FLT("gray_dumb",   "Grayscale converter (dumb version)",   xlg_flt_gray_dumb,   NULL,              0),
    FLT("gray",        "Grayscale convert (common version)",   xlg_flt_gray,        NULL,              1),
    FLT("hist_st",     "Histogram: Stretch",                   xlg_flt_hist_st,     NULL,              1),
    FLT("laplace",     "LaPlace",                              xlg_flt_laplace,     NULL,              1),
    FLT("matrix",      "Folding via matrix",                   xlg_flt_matrix,      NULL,              0),
    FLT("mirror",      "Flip: Horizontal",                     xlg_flt_mirror,      NULL,              1),
    FLT("negative",    "Negative image",                       xlg_flt_negative,    NULL,              1),
    FLT("noise_add",   "Add random noise",                     xlg_flt_noise_add,   NULL,              1),
    FLT("noise_reduc", "Noise reduction",                      xlg_flt_noise_reduc, NULL,              1),
    FLT("pinv",        "Partial inversion",                    xlg_flt_pinv,        NULL,              1),
    FLT("resize_nn",   "Resize: Nearest neighbor",             xlg_flt_resize_nn,   NULL,              1),
    FLT("resize_lx",   "Resize: Mono-linear X resample",       xlg_flt_resize_lx,   NULL,              1),
    FLT("resize_ly",   "Resize: Mono-linear Y resample",       xlg_flt_resize_ly,   NULL,              1),
    FLT("resize_bi",   "Resize: Bi-linear resample",           xlg_flt_resize_bi,   NULL,              1),
    FLT("robert",      "Find edges",                           xlg_flt_robert,      xlg_flt_robert_GS, 1),
    FLT("rol90",       "Rotate: Left by 90 degrees",           xlg_flt_rol90,       NULL,              0),
    FLT("ror90",       "Rotate: Right by 90 degrees",          xlg_flt_ror90,       NULL,              0),
    FLT("rotate",      "Rotate: Free rotation",                xlg_flt_rotate,      NULL,              1),
    FLT("solarize",    "Solarize (partial negation)",          xlg_flt_solarize,    NULL,              1),

    FLT("mx2_doom_soft", "Folding/Matrix: Soften: Doom transparency support", xlg_flt_matrix2_doom_soft, NULL, 0),
    FLT("mx2_soft",      "Folding/Matrix: Special soften for bounds",         xlg_flt_matrix2_soft,      NULL, 1),
    FLT(NULL, NULL, NULL, NULL, 0),
};
#undef FLT

static void strap_filters(void) {
    XLG_filter *i = fltinfo;
    if(pthread_mutex_trylock(&strap_mutex) < 0) { return; }
    ++strapped;
    while(i->name != NULL) {
        xlg_flt_register(i);
        ++i;
    }
    return;
}

//==[ End of file ]============================================================
