/*=============================================================================
libxlg
base/matrix.c - Control functions for matrices
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: matrix.c,v 1.7 2004/04/25 05:24:16 root Exp $
=============================================================================*/
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xlg.h"

static void strap_matrices(void);

static pthread_mutex_t strap_mutex = PTHREAD_MUTEX_INITIALIZER;
static int strapped = 0;

static pthread_mutex_t matrix_list_mutex = PTHREAD_MUTEX_INITIALIZER;
static XLG_matrix *matrix_list = NULL;
static size_t num_matrices = 0;

//-----------------------------------------------------------------------------
int xlg_mx_register(XLG_matrix *mp) {
    /* Registers a matrix within the xlg_mx database, which can later be
    retrieved using xlg_mx_list(). A matrix does not need to be enregistered,
    it can be used independently. xlg_mx_register() returns -errno upon error,
    0 if the matrix is already registered, and >0 for success. */
    XLG_matrix *lp = NULL, *travp;

    pthread_mutex_lock(&matrix_list_mutex);
    travp = (XLG_matrix *)&matrix_list;

    while(travp->Next != NULL) {
        // Alphabetic sorting
        int v = strcmp(mp->name, travp->Next->name);
        if(v == 0) {
            pthread_mutex_unlock(&matrix_list_mutex);
            return 0;
        }
        if(v < 0) {
            lp = travp->Next;
            break;
        }
        travp = travp->Next;
    }

    travp->Next = mp;
    mp->Next = lp;
    ++num_matrices;
    pthread_mutex_unlock(&matrix_list_mutex);
    return 1;
}

int xlg_mx_unregister(XLG_matrix *mp) {
    /* Unregisters a matrix. Returns 0 if there was no such matrix in the list,
    >0 if unregistering succeeded, and -errno for an error. */
    XLG_matrix *travp;

    pthread_mutex_lock(&matrix_list_mutex);
    travp = (XLG_matrix *)&matrix_list;
    while(travp->Next != NULL) {
        if(strcmp(travp->Next->name, mp->name) == 0) {
            travp->Next = mp->Next;
            --num_matrices;
            pthread_mutex_unlock(&matrix_list_mutex);
            mp->Next = NULL;
            return 1;
        }
        travp = travp->Next;
    }
    return 0;
}

int xlg_mx_list(XLG_matrix ***p) {
    /* xlg_mx_list() returns a list of available matrices. The list is
    allocated with malloc() and is stored in *p. The number of matrices is
    returned. If there are no matrices currently registered, nothing is
    allocated and 0 is returned. Also, nothing is done upon error, in which
    case -errno is returned.
    Applications may NOT rely on .Next, since this is only an internal field
    for the xlg_mx code. */
    XLG_matrix *travp, **area, **ap;
    int n;

    if(!strapped) { strap_matrices(); }
    pthread_mutex_lock(&matrix_list_mutex);
    if((n = num_matrices) == 0) { return 0; }
    if((area = ap = malloc(sizeof(XLG_matrix *) * n)) == NULL) {
        return -errno;
    }

    travp = matrix_list;
    while(n--) {
        *ap++ = travp;
        travp = travp->Next;
    }

    n = num_matrices;
    pthread_mutex_unlock(&matrix_list_mutex);
    *p = area;
    return n;
}

XLG_matrix *xlg_mx_get(const char *nm) {
    /* Returns the XLG_matrix structure for a given matrix. Returns NULL if
    no such matrix was found in the list. */
    XLG_matrix *travp;

    if(nm == NULL) { return NULL; }
    if(!strapped) { strap_matrices(); }
    pthread_mutex_lock(&matrix_list_mutex);
    travp = matrix_list;

    while(travp != NULL) {
        if(strcmp(travp->name, nm) == 0) {
            pthread_mutex_unlock(&matrix_list_mutex);
            return travp;
        }
        travp = travp->Next;
    }

    pthread_mutex_unlock(&matrix_list_mutex);
    return NULL;
}

//-----------------------------------------------------------------------------
static const double table_edge[] = {
     1, -1,  1,
    -1,  0, -1,
     1, -1,  1,
};
static const double table_gauss[] = {
    1, 2, 1,
    2, 4, 2,
    1, 2, 1,
};
static const double table_prewitt1[] = {
    -1, 0, 1,
    -1, 0, 1,
    -1, 0, 1,
};
static const double table_prewitt2[] = {
    -1, -1, -1,
     0,  0,  0,
     1,  1,  1,
};
static const double table_sharpen[] = {
    -1, -1, -1,
    -1, 12, -1,
    -1, -1, -1,
};
static const double table_sobel1[] = {
    -1, 0, 1,
    -2, 0, 2,
    -1, 0, 1,
};
static const double table_sobel2[] = {
    -1, -2, -1,
     0,  0,  0,
     1,  2,  1,
};
static const double table_soft[] = {
    1, 1, 1,
    1, 1, 1,
    1, 1, 1,
};
static const double table_unkn[] = {
    0, -1, 1,
    0, -1, 1,
    0, -1, 1,
};
static const double table_rgb2xyz[] = {
    0.412453, 0.357580, 0.180423,
    0.212671, 0.715160, 0.072169,
    0.019334, 0.119193, 0.950227,
};
static const double table_xyz2rgb[] = {
     3.240479, -1.537150, -0.498535,
    -0.969256,  1.875992,  0.041556,
     0.055648, -0.204043,  1.057311,
};
static const double table_rgb2yiq[] = {
    0.229,  0.587,  0.114,
    0.596, -0.275, -0.321,
    0.212, -0.523,  0.311,
};
static const double table_yiq2rgb[] = {
    1,  0.956,  0.621,
    1, -0.272, -0.647,
    1, -1.105,  1.702,
};
static const double table_psp8_blur[] = {
    1, 1, 2, 1, 1,
    1, 2, 2, 2, 1,
    2, 2, 8, 2, 2,
    1, 2, 2, 2, 1,
    1, 1, 2, 1, 1,
};
static const double table_psp8_blurM[] = {
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
};
static const double table_psp8_dvis[] = {
    -1, -2, -1, -2, -1,
     2, -1,  2, -1, -2,
    -1, -2,  5, -2, -1,
     2, -1,  2, -1,  2,
    -1, -2, -1, -2, -1,
};
static const double table_psp8_edgeE[] = {
    0,  1, 0,
    1, -5, 1,
    0,  1, 0,
};
static const double table_psp8_edgeEM[] = {
    -1, -1, -1,
    -1,  9, -1,
    -1, -1, -1,
};
static const double table_psp8_edgeEX[] = {
    -1, -1, -1,
     0,  1,  0,
    -1, -1, -1,
};
static const double table_psp8_edgeEY[] = {
    -1, 0, -1,
    -1, 1, -1,
    -1, 0, -1,
};
static const double table_psp8_edgeF[] = {
    -1, -1, -1,
    -1,  8, -1,
    -1, -1, -1,
};
static const double table_psp8_emboss[] = {
    -1, 0, 0,
     0, 0, 0,
     0, 0, 1,
};
static const double table_psp8_sharp[] = {
     0, -1,  0,
    -1, 10, -1,
     0, -1,  0,
};
static const double table_psp8_sharpMJ[] = {
    -1,  0, -1,  0, -1,
     0,  0, -1,  0,  0,
    -1, -1, 13, -1, -1,
     0,  0, -1,  0,  0,
    -1,  0, -1,  0, -1,
};
static const double table_psp8_sharpM[] = {
    -1, -1, -1, -1, -1,
    -1, -2, -2, -2, -1,
    -1, -2, 64, -2, -1,
    -1, -2, -2, -2, -1,
    -1, -1, -1, -1, -1,
};
static const double table_psp8_soften[] = {
     6, 12,  6,
    12, 25, 12,
     6, 12,  6,
};
static const double table_psp8_vibrate[] = {
     9, 0, 3, 0, 9,
     0, 0, 0, 0, 0,
     3, 0, 1, 0, 3,
     0, 0, 0, 0, 0,
     9, 0, 3, 0, 9,
};

#define MX(__name, __desc, __sx, __sy, __sz, __tbl, __div, __bias) \
 {NULL, (__name), (__desc), (__sx), (__sy), (__sz), (__tbl), (__div), (__bias)}
static XLG_matrix mxinfo[] = {
    MX("edge",    "Filter: Edge finder",          3, 3, 1, table_edge,     1,  0),
    MX("gauss",   "Filter: Gaussian blur",        3, 3, 1, table_gauss,    16, 0),
    MX("pw1",     "Filter: 1st Prewitt operator", 3, 3, 1, table_prewitt1, 1,  0),
    MX("pw2",     "Filter: 2nd Prewitt operator", 3, 3, 1, table_prewitt2, 1,  0),
    MX("sharpen", "Filter: Sharpen",              3, 3, 1, table_sharpen,  4,  0),
    MX("sob1",    "Filter: 1st Sobel operator",   3, 3, 1, table_sobel1,   1,  0),
    MX("sob2",    "Filter: 2nd Sobel operator",   3, 3, 1, table_sobel2,   1,  0),
    MX("soft",    "Filter: Soften",               3, 3, 1, table_soft,     9,  0),
    MX("unkn",    "Filter: Unnamed table",        3, 3, 1, table_unkn,     1,  0),

    MX("rgb2xyz", "RGB-to-XYZ conversion matrix", 3, 3, 1, table_rgb2xyz, 1, 0),
    MX("rgb2yiq", "RGB-to-YIQ conversion matrix", 3, 3, 1, table_rgb2yiq, 1, 0),
    MX("yiq2rgb", "YIQ-to-RGB conversion matrix", 3, 3, 1, table_yiq2rgb, 1, 0),
    MX("xyz2rgb", "XYZ-to-RGB conversion matrix", 3, 3, 1, table_xyz2rgb, 1, 0),

    MX("psp8_blur",   "PSP8: Blur",               5, 5, 1, table_psp8_blur,  44, 0),
    MX("psp8_blurM",  "PSP8: Blur more",          5, 5, 1, table_psp8_blurM, 25, 0),
    MX("psp8_dvis",   "PSP8: Double vision",      5, 5, 1, table_psp8_dvis,  -7, 0),
    MX("psp8_edgeE",  "PSP8: Edge enhance",       3, 3, 1, table_psp8_edgeE, -1, 0),
    MX("psp8_edgeEM", "PSP8: Edge enhance more",  3, 3, 1, table_psp8_edgeEM, 1, 0),
    MX("psp8_edgeEX", "PSP8: Edge enh. horiz.",   3, 3, 1, table_psp8_edgeEX, 1, 0),
    MX("psp8_edgeEY", "PSP8: Edge enh. vertical", 3, 3, 1, table_psp8_edgeEY, 1, 0),
    MX("psp8_edgeF",  "PSP8: Find edges",         3, 3, 1, table_psp8_edgeF,  1, 0),
    // psp_edgeFX: (PSP8) Find horizontal edges -> table_prewitt1
    // psp_edgeFY: (PSP8) Find vertical edges   -> table_prewitt2
    MX("psp8_emboss",  "PSP8: Emboss",        3, 3, 1, table_psp8_emboss,  1, 128),
    MX("psp8_sharp",   "PSP8: Sharpen",       3, 3, 1, table_psp8_sharp,   6, 0),
    MX("psp8_sharpMJ", "PSP8: Sharpen major", 5, 5, 1, table_psp8_sharpMJ, 1, 0),
    MX("psp8_sharpM",  "PSP8: Sharpen more",  5, 5, 1, table_psp8_sharpM, 32, 0),
    MX("psp8_soften",  "PSP8: Soften",        3, 3, 1, table_psp8_soften, 97, 0),
    // psp8_softenM: (PSP8) Soften more -> table_soft
    MX("psp8_trace",   "PSP8: Trace Contour", 3, 3, 1, table_psp8_edgeEM,   1, 255),
    MX("psp8_vibrate", "PSP8: Vibrate",       5, 5, 1, table_psp8_vibrate, 49, 0),
    MX(NULL, NULL, 0, 0, 0, NULL, 0.0, 0.0),
};
#undef MX

static void strap_matrices(void) {
    XLG_matrix *i = mxinfo;
    if(pthread_mutex_trylock(&strap_mutex) < 0) { return; }
    ++strapped;
    while(i->name != NULL) {
        xlg_mx_register(i);
        ++i;
    }
    return;
}

//==[ End of file ]============================================================
