/*=============================================================================
libxlg
base/pic.c - Functions for dealing with simple pictures,
  to show the use of filters and matrices
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: pic.c,v 1.5 2004/04/15 13:54:40 root Exp $
=============================================================================*/
#include <sys/stat.h>
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "xlg.h"
#define MAXLNLEN 512

//-----------------------------------------------------------------------------
XLG_pic *xlg_pic_open(const char *file) {
    char buf[MAXLNLEN];
    struct stat sb;
    XLG_pic *ret;
    FILE *fp;
    int eax;

    if((ret = malloc(sizeof(XLG_pic))) == NULL) { return NULL; }
    if(strcmp(file, "-") == 0) { fp = fopen("/dev/stdin", "rb"); }
    else { fp = fopen(file, "rb"); }

    if(fp == NULL) {
        eax = errno;
        xlg_pic_free(ret);
        errno = eax;
        return NULL;
    }

    fgets(buf, MAXLNLEN, fp);
    if(memcmp(buf, "P6\n", 3) != 0) {
        fclose(fp);
        xlg_pic_free(ret);
        errno = 0;
        return NULL;
    }

    do {
        fgets(buf, MAXLNLEN, fp);
    } while(*buf == '#');

    sscanf(buf, "%d %d", &ret->sx, &ret->sy);
    fgets(buf, MAXLNLEN, fp);

    fstat(fileno(fp), &sb);
    ret->dsize = sb.st_size - ftell(fp);

    if((ret->data = malloc(ret->dsize)) == NULL) {
        eax = errno;
        xlg_pic_free(ret);
        errno = eax;
        return NULL;
    }

    fread(ret->data, ret->dsize, 1, fp);
    fclose(fp);
    return ret;
}

XLG_pic *xlg_pic_new(int x, int y) {
    XLG_pic *ret;
    int eax;

    if((ret = malloc(sizeof(XLG_pic))) == NULL) { return NULL; }
    ret->sx = x;
    ret->sy = y;
    ret->dsize = x * y * 3;

    if((ret->data = malloc(ret->dsize)) == NULL) {
        eax = errno;
        xlg_pic_free(ret);
        errno = eax;
        return NULL;
    }

    xlg_pic_zero(ret);
    return ret;
}

XLG_pic *xlg_pic_dup(XLG_pic *src) {
    XLG_pic *ret;
    int eax;

    if((ret = malloc(sizeof(XLG_pic))) == NULL) { return NULL; }
    ret->sx = src->sx;
    ret->sy = src->sy;
    ret->dsize = src->dsize;

    if((ret->data = malloc(ret->dsize)) == NULL) {
        eax = errno;
        xlg_pic_free(ret);
        errno = eax;
        return NULL;
    }

    memcpy(ret->data, src->data, ret->dsize);
    return ret;
}

XLG_pic *xlg_pic_zero(XLG_pic *p) {
    // Clears the image data
    memset(p->data, 0, p->dsize);
    return p;
}

XLG_pic *xlg_pic_border(XLG_pic *p) {
    int x, y;
    for(x = 0; x < p->sx; ++x) {
        xlg_pic_setpx_sw(p, x, 0, 0x80);
        xlg_pic_setpx_sw(p, x, p->sy - 1, 0x80);
    }
    for(y = 0; y < p->sy; ++y) {
        xlg_pic_setpx_sw(p, 0, y, 0x80);
        xlg_pic_setpx_sw(p, p->sx - 1, y, 0x80);
    }
    return p;
}

uint8_t *xlg_pic_getpx(XLG_pic *p, int x, int y) {
    // Gets a pixel, returning NULL if x and y are out of bounds
    if(x < 0 || x >= p->sx || y < 0 || y >= p->sy) { return NULL; }
    return &p->data[(y * p->sx + x) * 3];
}

uint8_t *xlg_pic_getpxn(XLG_pic *p, int x, int y) {
    // Gets a pixel, returning a black pixel if x and y are out of bounds
    static uint8_t zero[3] = {0, 0, 0};
    if(x < 0 || x >= p->sx || y < 0 || y >= p->sy) { return zero; }
    return &p->data[(y * p->sx + x) * 3];
}

uint8_t xlg_pic_getpx_sw(XLG_pic *p, int x, int y) {
    // Gets the gray value of a pixel, or 0 if x or y are out of bounds
    uint8_t *q = &p->data[(y * p->sx + x) * 3];
    if(x < 0 || x >= p->sx || y < 0 || y >= p->sy) { return 0; }

    /* The original grayscale algorithm was (77r + 151g + 26g) / 254 but I
    adjusted it for speed and hoping that the results are [mostly] the same. */
    return (78 * q[0] + 151 * q[1] + 27 * q[2]) >> 8;
}

uint8_t *xlg_pic_setpx(XLG_pic *p, int x, int y, uint8_t r, uint8_t g, uint8_t b) {
    // Sets a pixel, taking RGB values as args
    uint8_t *q = &p->data[(y * p->sx + x) * 3];
    if(x < 0 || x >= p->sx || y < 0 || y >= p->sy) { return NULL; }
    q[0] = XLG_PV_SQUASH(r);
    q[1] = XLG_PV_SQUASH(g);
    q[2] = XLG_PV_SQUASH(b);
    return q;
}

uint8_t *xlg_pic_setpxp(XLG_pic *p, int x, int y, uint8_t *c) {
    // Sets a pixel, taking RGB values as a pointer
    uint8_t *q = &p->data[(y * p->sx + x) * 3];
    if(x < 0 || x >= p->sx || y < 0 || y >= p->sy) { return NULL; }
    q[0] = XLG_PV_SQUASH(c[0]);
    q[1] = XLG_PV_SQUASH(c[1]);
    q[2] = XLG_PV_SQUASH(c[2]);
    return q;
}

uint8_t xlg_pic_setpx_sw(XLG_pic *p, int x, int y, uint8_t g) {
    uint8_t *q = &p->data[(y * p->sx + x) * 3];
    if(x < 0 || x >= p->sx || y < 0 || y >= p->sy) { return 0; }
    q[0] = q[1] = q[2] = XLG_PV_SQUASH(g);
    return g;
}

int xlg_pic_store(XLG_pic *p, const char *file) {
    FILE *fp;

    if(strcmp(file, "-") == 0) { fp = fopen("/dev/stdout", "wb"); }
    else { fp = fopen(file, "wb"); }
    if(fp == NULL) { return 0; }

    fprintf(fp, "P6\n%d %d\n255\n", p->sx, p->sy);
    fwrite(p->data, p->dsize, 1, fp);
    fclose(fp);
    return 1;
}

void xlg_pic_free(XLG_pic *p) {
    if(p->data != NULL) { free(p->data); }
    free(p);
    return;
}

//==[ End of file ]============================================================
