/*=============================================================================
libxlg
filters/dft.c - Discrete Fourier Transforms
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: dft.c,v 1.8 2004/04/05 20:10:43 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xlg.h"
#include "filters/strap.h"

const double m_pi2 = M_PI * 2;

//-----------------------------------------------------------------------------
int xlg_flt_dft(XLG_pic *ip, XLG_pic **opp, int argc, char **argv) {
    int C, M = ip->sy, N = ip->sx; // channel
    double *c, *s;
    XLG_pic *op;

    if((op = *opp = xlg_pic_new(ip->sx, ip->sy)) == NULL) { return -errno; }
    setvbuf(stderr, NULL, _IONBF, 0);
    c = malloc(M * N * sizeof(double));
    s = malloc(M * N * sizeof(double));

    for(C = 0; C < 3; ++C) {
        int u, v, x, y;
        memset(c, 0, M * N * sizeof(double));
        memset(s, 0, M * N * sizeof(double));

        for(v = 0; v < M; ++v) {
            for(u = 0; u < N; ++u) {
                double s_arg = 1 / sqrt(N * M);
                fprintf(stderr, "\r\e[2K" "Doing DFT C=%d v=%3d u=%3d",
                 C, v, u);

                for(y = 0; y < M; ++y) {
                    for(x = 0; x < N; ++x) {
                        double w_arg = -m_pi2 * (x * u / N + y * v / M);
                        uint8_t px = xlg_pic_getpxn(ip, x, y)[C];
                        c[N*v+u] += px * cos(w_arg);
                        s[N*v+u] += px * sin(w_arg);
                    }
                }

                c[N*v+u] *= s_arg;
                s[N*v+u] *= s_arg;
            }
        }

        // kill freq
        for(v = M / 2; v < M; ++v) {
            for(u = N / 2; u < N; ++u) {
                c[N*v+u] = s[N*v+u] = 0;
            }
        }

        // IDFT back transformation
        for(y = 0; y < M; ++y) {
            for(x = 0; x < N; ++x) {
                double r = 0;
                fprintf(stderr, "\r\e[2K" "Doing IDFT C=%d y=%3d x=%3d",
                 C, y, x);

                for(v = 0; v < M; ++v) {
                    for(u = 0; u < N; ++u) {
                        double w_arg = m_pi2 * (x * u / N + y * v / M);
                        r += c[N*v+u] * cos(w_arg);
                        r -= s[N*v+u] * sin(w_arg);
                    }
                }

                r *= 1 / sqrt(N * M);
                xlg_pic_getpx(op, x, y)[C] = XLG_PV_SQUASH(r);
            }
        }
    } // channel rerun

    free(c);
    free(s);
    fprintf(stderr, "\n");
    return 1;
}

int xlg_flt_dft_GS(XLG_pic *ip, XLG_pic **opp, int argc, char **argv) {
    int N = ip->sx, M = ip->sy;
    int u, v, x, y;
    double *c, *s;
    XLG_pic *op;

    if((op = *opp = xlg_pic_new(ip->sx, ip->sy)) == NULL) { return -errno; }
    setvbuf(stderr, NULL, _IONBF, 0);
    c = calloc(M * N, sizeof(double));
    s = calloc(M * N, sizeof(double));

    for(v = 0; v < M; ++v) {
        for(u = 0; u < N; ++u) {
            double s_arg = 1 / sqrt(N * M);
            fprintf(stderr, "\r\e[2K" "Doing DFT v=%3d u=%3d", v, u);

            for(y = 0; y < M; ++y) {
                for(x = 0; x < N; ++x) {
                    double w_arg = -m_pi2 * (x * u / N + y * v / M);
                    uint8_t px = xlg_pic_getpx_sw(ip, x, y);
                    c[N*v+u] += px * cos(w_arg);
                    s[N*v+u] += px * sin(w_arg);
                }
            }

            c[N*v+u] *= s_arg;
            s[N*v+u] *= s_arg;
        }
    }

    // kill freq
    for(v = M / 2; v < M; ++v) {
        for(u = N / 2; u < N; ++u) {
            c[N*v+u] = s[N*v+u] = 0;
        }
    }

    // IDFT back transformation
    for(y = 0; y < M; ++y) {
        for(x = 0; x < N; ++x) {
            double r = 0;
            fprintf(stderr, "\r\e[2K" "Doing IDFT y=%3d x=%3d", y, x);

            for(v = 0; v < M; ++v) {
                for(u = 0; u < N; ++u) {
                    double w_arg = m_pi2 * (x * u / N + y * v / M);
                    r += c[N*v+u] * cos(w_arg);
                    r -= s[N*v+u] * sin(w_arg);
                }
            }

            r *= 1 / sqrt(N * M);
            xlg_pic_setpx_sw(op, x, y, XLG_PV_SQUASH(r));
        }
    }

    free(c);
    free(s);
    fprintf(stderr, "\n");
    return 1;
}

//==[ End of file ]============================================================
