/*=============================================================================
libxlg
run.c - General XLG wrapper
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: run.c,v 1.5 2004/03/28 11:58:28 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xlg.h"

static void list_filters(void);
static void list_matrices(void);
static int get_options(int *argc, const char ***argv);

static char *infile = "coke.ppm", *outfile = "out.ppm", *filter_s = NULL,
 *matrix_s = NULL;
static int list_all = 0, list_it = 0, use_gs = 0;

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    XLG_pic *d_in, *d_out;
    XLG_filter *flt;
    int eax;

    if(!get_options(&argc, (const char ***)&argv)) { return !0; }

    if(list_it) {
        list_filters();
        list_matrices();
        return EXIT_SUCCESS;
    }

    if((d_in = xlg_pic_open(infile)) == NULL) {
        perror("xlg_pic_open()");
        return EXIT_FAILURE;
    }

    if(matrix_s != NULL) {
        char *args[] = {matrix_s, NULL};
        if((flt = xlg_flt_get("matrix")) == NULL) {
            fprintf(stderr, "Matrix filter not found (errno=%d)\n", errno);
            return EXIT_FAILURE;
        }

        if(use_gs && flt->proc_gs != NULL) {
            eax = flt->proc_gs(d_in, &d_out, 1, args);
        } else {
            eax = flt->proc(d_in, &d_out, 1, args);
        }

        if(eax < 0) {
            perror("xlg_flt_matrix()");
            return EXIT_FAILURE;
        } else if(eax == 0) {
            return EXIT_FAILURE;
        }
    } else {
        if((flt = xlg_flt_get(filter_s)) == NULL) {
            fprintf(stderr, "No such filter \"%s\"\n", filter_s);
            return EXIT_FAILURE;
        }

        if(use_gs && flt->proc_gs != NULL) {
            eax = flt->proc_gs(d_in, &d_out, argc, argv);
        } else {
            eax = flt->proc(d_in, &d_out, argc, argv);
        }

        if(eax < 0) {
            perror("flt->proc()");
            return EXIT_FAILURE;
        } else if(eax == 0) {
            return EXIT_FAILURE;
        }
    }

    xlg_pic_store(d_out, outfile);
    xlg_pic_free(d_in);
    xlg_pic_free(d_out);
    return EXIT_SUCCESS;
}

static void list_filters(void) {
    XLG_filter **flt, **fltp;
    size_t n = xlg_flt_list(&flt), hidden = 0;
    fltp = flt;

    if(n < 0) {
        perror("xlg_flt_list()");
        return;
    }

    if(n == 0) {
        printf(">>> No filters currently registered in libxlg\n");
        return;
    }

    printf(">>> %d filters currently registered in libxlg:\n", n);
    while(n--) {
        if((**fltp).show) {
            printf("  %-15s   (%s)\n", (**fltp).name, (**fltp).desc);
        } else {
            if(list_all) {
                printf("  %-15s * (%s)\n", (**fltp).name, (**fltp).desc);
            }
            ++hidden;
        }
        ++fltp;
    }
    printf(">>> %d of which are hidden\n", hidden);

    free(flt);
    return;
}     

static void list_matrices(void) {
    XLG_matrix **m, **mp;
    size_t n = xlg_mx_list(&m);
    mp = m;

    if(n < 0) {
        perror("xlg_mx_list()");
        return;
    }

    if(n == 0) {
        printf(">>> No matrices currently registered in libxlg\n");
        return;
    }

    printf(">>> %d matrices currently registered in libxlg:\n", n);
    while(n--) {
        printf("  %-15s   (%s)\n", (**mp).name, (**mp).desc);
        ++mp;
    }

    free(m);
    return;
}

//-----------------------------------------------------------------------------
static int count_args(const char **args) {
    int argk = 0;
    while(*args++ != NULL) { ++argk; }
    return argk;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    struct poptOption options_table[] = {
        {NULL, 'A', POPT_ARG_NONE, &list_all, 0, "List everything", NULL},
        {NULL, 'F', POPT_ARG_STRING, &filter_s, 0, "Filter to apply", NULL},
        {NULL, 'G', POPT_ARG_NONE, &use_gs, 0, "Use grayscale variant", NULL},
        {NULL, 'L', POPT_ARG_NONE, &list_it, 0, "List filters and matrices", NULL},
        {NULL, 'M', POPT_ARG_STRING, &matrix_s, 0, "Matrix to apply (implies -F matrix)", NULL},
        {NULL, 'i', POPT_ARG_STRING, &infile, 0, "Input file", NULL},
        {NULL, 'o', POPT_ARG_STRING, &outfile, 0, "Output file", NULL},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0);
    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    poptFreeContext(ctx);
    return 1;
}

//==[ End of file ]============================================================
